START = "START"
FINALIZE = "FINALIZE"
PROCESS_BUCKET = "PROCESS_BUCKET"
GENERATE_BUCKET = "GENERATE_BUCKET"
CONSUME_BUCKET = "CONSUME_BUCKET"
PRODUCE_BUCKET = "PRODUCE_BUCKET"
PRODUCE_TIMING = "PRODUCE_TIMING"
UPDATE_COUNTER = "UPDATE_COUNTER"

NODE_ACTIONS = [
    START,
    FINALIZE,
    CONSUME_BUCKET,
    PRODUCE_BUCKET,
    PROCESS_BUCKET,
    GENERATE_BUCKET,
    PRODUCE_TIMING,
    UPDATE_COUNTER,
]
